package javato.activetesting;

import java.util.* ;
import java.io.* ;

import javato.activetesting.common.Parameters ;
import javato.activetesting.activechecker.ActiveChecker;
import javato.activetesting.analysis.AnalysisImpl;
import javato.activetesting.reentrant.IgnoreRentrantLock;

import javato.activetesting.coverage.* ;

/**
 * Copyright (c) 2007-2008,
 * Koushik Sen    <ksen@cs.berkeley.edu>
 * All rights reserved.
 * <p/>
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * <p/>
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 * <p/>
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 * <p/>
 * 3. The names of the contributors may not be used to endorse or promote
 * products derived from this software without specific prior written
 * permission.
 * <p/>
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

public class PredictAnalysis extends AnalysisImpl {

	class IntPair {
		public Integer x ;
		public Integer y ;

		public IntPair(Integer x, Integer y) {
			this.x = new Integer(x.intValue()) ;
			this.y = new Integer(y.intValue()) ;
		}
	}

	public static LinkedHashMap<Integer, LinkedList<CovEvent>> execution =
		new LinkedHashMap<Integer, LinkedList<CovEvent>>() ;
	/* execution: Thread -> a sequence of CovEvent */

	public static LinkedHashSet<SyncPair> coverable = new LinkedHashSet<SyncPair>() ;

	public void initialize() {
	}

	public void lockBefore(Integer iid, Integer thread, Integer lock) {
		synchronized (execution) {
			getThreadModel(thread).addLast(new SyncEvent(CovEvent.LOCK, iid, lock)) ;
		}
	}

	public void lockAfter(Integer iid, Integer thread, Integer lock) {
	}

	public void unlockAfter(Integer iid, Integer thread, Integer lock) {
		synchronized (execution) {
			getThreadModel(thread).addLast(new SyncEvent(CovEvent.UNLOCK, iid, lock)) ;	    
		}
	}

	public void newExprAfter(Integer iid, Integer object, Integer objOnWhichMethodIsInvoked) {
	}

	public void methodEnterBefore(Integer iid) {
	}

	public void methodExitAfter(Integer iid) {
	}

	public void startBefore(Integer iid, Integer parent, Integer child) {
		synchronized (execution) {
			getThreadModel(parent).addLast(new ThreadEvent(CovEvent.START, iid, child)) ;
		}
	}

	public void waitAfter(Integer iid, Integer thread, Integer lock) {		
		synchronized (execution) {
			getThreadModel(thread).addLast(new SyncEvent(CovEvent.WAIT, iid, lock)) ;
		}
	}

	public void notifyBefore(Integer iid, Integer thread, Integer lock) {
		synchronized (execution) {
			getThreadModel(thread).addLast(new SyncEvent(CovEvent.NOTIFY, iid, lock)) ;
		}		
	}

	public void notifyAllBefore(Integer iid, Integer thread, Integer lock) {
		synchronized (execution) {
			getThreadModel(thread).addLast(new SyncEvent(CovEvent.NOTIFYALL, iid, lock)) ;
		}
	}

	public void joinAfter(Integer iid, Integer parent, Integer child) {
	}

	public void readBefore(Integer iid, Integer thread, Long memory) {
		/*		
		   synchronized (execution) {
		   getThreadModel(thread).addLast(new RWEvent(CovEvent.READ, iid, memory)) ;
		   }
		*/
		   
	}
	public void readAfter(Integer iid, Integer thread, Long memory) {
	}

	public void writeBefore(Integer iid, Integer thread, Long memory) {
		/*
		   synchronized (execution) {
		   getThreadModel(thread).addLast(new RWEvent(CovEvent.WRITE, iid, memory)) ;
		   }
		 */
		   
	}
	public void writeAfter(Integer iid, Integer thread, Long memory) {
	}    

	public static boolean exclusive(LinkedList<Integer> s1, LinkedList<Integer> s2) {
		Iterator<Integer> itr = s1.iterator() ;
		while (itr.hasNext()) {
			if (s2.contains(itr.next()) == true)
				return false ;
		}
		return true ;
	}

	public static LinkedList<Integer> conjunction(LinkedList<Integer> s1, LinkedList<Integer> s2) {
		LinkedList<Integer> result = new LinkedList<Integer>() ;
		Iterator<Integer> itr ;
		for (itr = s1.iterator() ; itr.hasNext() == true; ) {
			int l = itr.next().intValue() ;
			if (s2.contains(new Integer(l)) == true)
				result.add(new Integer(l)) ;
		}
		return result ;
	}

	public void predictCoverableSyncPair() {
		Set<Integer> threadIDs = execution.keySet() ;
		int nThreads = threadIDs.size() ;
		CovEvent [][] events = null ;
		HashMap<Integer, Integer> threadOrder
			= new HashMap<Integer, Integer>() ; /* Child Thread Identifier -> Parent Thread Identifier */
		HashMap<Integer, Integer> mapThreadID = new HashMap<Integer, Integer>() ; /* thread id -> index */
		HashMap<Integer, Integer> mapIndexToThreadID = new HashMap<Integer, Integer>() ;

		events = new CovEvent[threadIDs.size()][] ;

		Iterator<Integer> itr = threadIDs.iterator() ;
		for (int i = 0 ; i < threadIDs.size() ; i++) {
			int threadID = itr.next() ;
			mapThreadID.put(threadID, i) ;
			mapIndexToThreadID.put(i, threadID) ;
			LinkedList<CovEvent> seq = execution.get(threadID) ;			
			events[i] = new CovEvent[seq.size()] ;
			for (int j = 0 ; j < seq.size() ; j++) {
				events[i][j] = seq.get(j) ;
				if (events[i][j].type == CovEvent.START) {
					ThreadEvent s = (ThreadEvent)events[i][j] ;
					threadOrder.put(s.thread, threadID) ;
				}
			}
		}

		for (int th1 = 0 ; th1 < nThreads ; th1++) {
			for (int th2 = 0 ; th2 < nThreads ; th2++) {
				if (th1 == th2) {
					continue ;
				}
				else {
					System.out.print("pair: ") ;
					System.out.print(mapIndexToThreadID.get(th1)) ;
					System.out.print(",") ;
					System.out.println(mapIndexToThreadID.get(th2)) ;
				}
				int i_size = events[th1].length ;
				int j_size = events[th2].length ;

				int i_begin = 0 ;
				int j_begin = 0 ;

				for (int i = 0 ; i < i_size ; i++) {
					if (events[th1][i].type == CovEvent.START) {
						ThreadEvent te = (ThreadEvent)events[th1][i] ;
						if (te.thread.equals(mapIndexToThreadID.get(th2)) == true) {
							i_begin = i + 1 ;
							break ;
						}
					}
				}
				if (i_begin == 0) {
					for (int j = 0 ; j < j_size ; j++) {
						if (events[th2][j].type == CovEvent.START) {
							ThreadEvent te = (ThreadEvent)events[th2][j] ;
							if (te.thread.equals(mapIndexToThreadID.get(th1)) == true) {
								j_begin = j + 1 ;
								break ;
							}
						}
					}				
				}

				LinkedList<Integer> LSi = new LinkedList<Integer>() ;
				HashMap<Integer, Integer> lastLock = new HashMap<Integer, Integer>() ; /* lock -> iid */
				HashMap<Integer, Integer> lastLockIndex = new HashMap<Integer, Integer>() ; /* lock -> index */
				HashMap<Long, Integer> lastWrite = new HashMap<Long, Integer>() ; /* memory -> iid */
				for (int i = 0 ; i < i_size ; i++) {			
					/* Add predictions */
					if (i >= i_begin) {
						LinkedList<Integer> LSj = new LinkedList<Integer>() ;
						for (int j = 0 ; j < j_size ; j++) {
							/* Add predictions */							
							if (j >= j_begin) {
								if (exclusive(LSi, LSj) == true) {
									if (events[th2][j].type == CovEvent.LOCK) {
										if (LSj.contains(events[th2][j]) != true) { /* ignore recursive locking case */
											SyncEvent se = (SyncEvent)events[th2][j] ;
											Integer lastLockIID = lastLock.get(se.lock) ;
											if (lastLockIID == null) {
												if (threadOrder.get(mapIndexToThreadID.get(th1)) == null) 
													coverable.add(new SyncPair(new Integer(-1), se.iid)) ;
											}
											else {
												coverable.add(new SyncPair(lastLockIID, se.iid)) ;
												if (i_begin <= lastLockIndex.get(se.lock).intValue()) {
													coverable.add(new SyncPair(se.iid, lastLockIID)) ;
												}
											}
										}
									}
									else if (events[th2][j].type == CovEvent.READ) {
										RWEvent re = (RWEvent)events[th2][j] ;
										Integer lastWriteIID = lastWrite.get(re.memory) ;
										coverable.add(new SyncPair(lastWriteIID, re.iid)) ;									
									}
									else if (events[th2][j].type == CovEvent.WRITE) {
										RWEvent we = (RWEvent)events[th2][j] ;
										Integer lastWriteIID = lastWrite.get(we.memory) ;
										coverable.add(new SyncPair(lastWriteIID, we.iid)) ;
										if (events[th1][i].type == CovEvent.READ) {
											RWEvent re = (RWEvent)events[th1][i] ;
											if (re.memory.longValue() == we.memory.longValue()) {
												coverable.add(new SyncPair(we.iid, re.iid)) ;
											}
										}
									}
									else {
										/* Currently, the others are not handled */
									}
								}
								else {
									if ((events[th1][i].type == CovEvent.LOCK) && (events[th2][j].type == CovEvent.LOCK)) {
										SyncEvent seth1 = (SyncEvent)events[th1][i] ;
										SyncEvent seth2 = (SyncEvent)events[th2][j] ;

										if (seth1.lock.intValue() == seth2.lock.intValue()) {

											LinkedList<Integer> commonLS = conjunction(LSi, LSj) ;
											for (Iterator<Integer> iconj = commonLS.iterator() ; iconj.hasNext() == true ; ) {
												int cLock = iconj.next() ;

												int th1_cLock_unlock = -1 ;
												int th1_next_acquire = -1 ;
												int th2_cLock_lock = -1 ;
												int th2_prev_acquire = -1 ;
											
												// Lookahead
												// 	For th1, find the index where cLock is unlockecd after i
												for (int th1_i = i + 1 ; th1_i < i_size ; th1_i++) {
													if (events[th1][th1_i].type == CovEvent.UNLOCK) {
														SyncEvent sei = (SyncEvent)events[th1][th1_i] ;
														if (sei.lock.intValue() == cLock) {
															th1_cLock_unlock = th1_i ;
															break ;
														}
													}
												}
												// 		 find the index that acquire i.operand after i
												for (int th1_i = i + 1 ; th1_i < i_size ; th1_i++) {
													if ((events[th1][th1_i].type == CovEvent.LOCK)) {
														SyncEvent sei = (SyncEvent)events[th1][th1_i] ;	
											    			if (sei.lock.intValue() == seth1.lock.intValue()) {
															th1_next_acquire = th1_i ;
															break ;
														}
													}
												}
												// 	For th2, find the index where cLock is hold before j
												for (int th2_i = j - 1 ; th2_i >= 0 ; th2_i--) {
													if (events[th2][th2_i].type == CovEvent.LOCK) {
														SyncEvent sei = (SyncEvent)events[th2][th2_i] ;
														if (sei.lock.intValue() == cLock) {
															th2_cLock_lock = th2_i ;
															break ;
														}
													}
												}
												// 		 find the index that acquire j.operand before j
												for (int th2_i = j - 1 ; th2_i >= 0 ; th2_i--) {
													if (events[th2][th2_i].type == CovEvent.LOCK) {
														SyncEvent sei = (SyncEvent)events[th2][th2_i] ;
													    	if (sei.lock.intValue() == seth2.lock.intValue()) {
															th2_prev_acquire = th2_i ;
															break ;
														}
													}
												} 		 

												if ((th1_next_acquire == -1) || (th1_cLock_unlock < th1_next_acquire)) {
													if ((th2_prev_acquire == -1) || (th2_prev_acquire < th2_cLock_lock)) {
														coverable.add(new SyncPair(seth1.iid, seth2.iid)) ;
													}
												}

											}
										}
									}
								}
							}
							/* Update context */						
							if (events[th2][j].type == CovEvent.LOCK) {
								SyncEvent se = (SyncEvent)events[th2][j] ;
								LSj.add(se.lock) ;
							}
							else if (events[th2][j].type == CovEvent.UNLOCK) {
								SyncEvent se = (SyncEvent)events[th2][j] ;								
								LSj.remove(se.lock) ;
							}							
						}
					}
					/* Update context */
					if (events[th1][i].type == CovEvent.LOCK) {
						SyncEvent se = (SyncEvent)events[th1][i] ;						
						Integer lastLockIID = lastLock.get(se.lock) ;
						if (lastLockIID == null) {
							if (threadOrder.get(mapIndexToThreadID.get(th1)) == null)
								coverable.add(new SyncPair(new Integer(-1), se.iid)) ;
						}
						else {
							coverable.add(new SyncPair(lastLockIID, se.iid)) ;
						}
						if (LSi.contains(se.lock) == false) {
							lastLock.put(se.lock, se.iid) ;
							lastLockIndex.put(se.lock, i) ;
						}
						LSi.add(se.lock) ;
					}
					else if (events[th1][i].type == CovEvent.UNLOCK) {
						SyncEvent se = (SyncEvent)events[th1][i] ;
						LSi.remove(se.lock) ;
					}
					else if (events[th1][i].type == CovEvent.WRITE) {
						RWEvent we = (RWEvent)events[th1][i] ;
						coverable.add(new SyncPair(lastWrite.get(we.memory), we.iid)) ;
						lastWrite.put(we.memory, we.iid) ;
					}

				}
			}
		}
	}

	private int getThreadOrder(Integer a, Integer b) {

		return 0 ;
	}

	public void printExecutionModel() {
		Set<Integer> tset = (Set<Integer>)execution.keySet() ;
		Iterator titr = tset.iterator() ;
		Integer [] threads = new Integer[tset.size()] ;

		for (int k = 0 ; titr.hasNext() == true ; k++)
			threads[k] = (Integer) titr.next() ;


		System.out.println("Concurrent execution model") ;
		System.out.println("--------------------------") ;
		for (int i = 0 ; i < threads.length ; i++) {
			LinkedList<CovEvent> seq = (LinkedList<CovEvent>)execution.get(threads[i]) ;

			System.out.println("Thread-" + threads[i] + ":") ;
			System.out.print("  ") ;
			for (int j = 0 ; j < seq.size() ; j++)
				System.out.print(seq.get(j) + " ") ;
			System.out.println("") ;
		}
		System.out.println("==========================") ;
		for (Iterator<SyncPair> itr = coverable.iterator() ; itr.hasNext() == true ; ) {
			System.out.print(itr.next() + " ") ;
		}
		System.out.println("") ;
	}

	public void finish() {
		/* in order to protocol of test driver */
		ObjectOutputStream out ;

		LinkedHashSet<SyncPair> s = new LinkedHashSet<SyncPair>() ;

		predictCoverableSyncPair() ;

		System.out.println("coverable items: " + coverable.size()) ;

		printExecutionModel() ;

		Parameters.writeIntegerList(Parameters.ERROR_LIST_FILE, 1) ;
		try {
			out = new ObjectOutputStream(new BufferedOutputStream(new FileOutputStream(Parameters.ERROR_LOG_FILE))) ;
			out.writeObject(s) ; 		/* used for metric of already covered items */
			out.writeObject(coverable) ; 	/* used for metric of not yet covered items */
			out.close() ;
		}	    
		catch (IOException e) {
			e.printStackTrace() ;
		}
	}

	private LinkedList<CovEvent> getThreadModel(Integer thread) {
		LinkedList<CovEvent> sequence = execution.get(thread) ;
		if (sequence == null) {
			execution.put(thread, new LinkedList<CovEvent>()) ;
			sequence = execution.get(thread) ;
		}
		return sequence ;
	}
}

