package javato.activetesting.scheduler;

import javato.activetesting.common.Parameters;
import javato.activetesting.activechecker.ActiveChecker;

/**
 * Copyright (c) 2007-2008,
 * Koushik Sen    <ksen@cs.berkeley.edu>
 * All rights reserved.
 * <p/>
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * <p/>
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 * <p/>
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 * <p/>
 * 3. The names of the contributors may not be used to endorse or promote
 * products derived from this software without specific prior written
 * permission.
 * <p/>
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

public class OneMinuteBreaker extends Thread {
	public OneMinuteBreaker() {
		super("ipc3");
		setDaemon(true);
		setPriority(Thread.MIN_PRIORITY);
	}
	
	public void run() {
		Object lock = new Object();
		long time = System.currentTimeMillis();
		synchronized (lock) {
			try {
				while(System.currentTimeMillis()-time < 60 * 1000){
					lock.wait(Parameters.stallCheckerInterval);
				}
				
				shutdown();
			}catch (InterruptedException e){
			}
		}
	}
	
	public static void shutdown(){
		System.out.println("----------------------------------------------------------------------\n(debug) OneMinuteBreaker.shutdown()\n");
		Thread[] tList = new Thread[Thread.activeCount()];
        int numThreads = Thread.enumerate(tList);
        for (int i = 0; i<numThreads; i++){
        	System.out.println("---Stacktrace for thread Name: "+tList[i].getName()+", ID: "+tList[i].getId());
        	for (StackTraceElement e : tList[i].getStackTrace()){
        		System.out.println(e);
        	}
        }
		System.exit(0);
	}
}
